﻿using MindFusion.Diagramming;
using MindFusion.Drawing;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;
using Xamarin.Forms;

using Brush = MindFusion.Drawing.Brush;
using LinearGradientBrush = MindFusion.Drawing.LinearGradientBrush;

namespace DecisionLayout
{
	// Learn more about making custom code visible in the Xamarin.Forms previewer
	// by visiting https://aka.ms/xamarinforms-previewer
	[DesignTimeVisible(false)]
	public partial class MainPage : ContentPage
	{
		public MainPage()
		{
			InitializeComponent();

			diagram = diagramView.Diagram;
			diagram.Bounds = new Rectangle(0, 0, 2000, 2000);

			terminalBrush = new LinearGradientBrush(
				Color.FromRgb(0xFF, 0xFF, 0xC0), Color.FromRgb(0xBF, 0xBF, 0x90), 90);
			decisionBrush = new LinearGradientBrush(
				Color.FromRgb(0xF0, 0xFF, 0xFF), Color.FromRgb(0x7F, 0xDF, 0x8F), 90);
			normalBrush = new LinearGradientBrush(
				Color.FromRgb(0xE0, 0xFF, 0xFF), Color.FromRgb(0xA8, 0xBF, 0xBF), 90);
			terminalPen = new Pen(Color.Black);
			terminalPen.Width = 0.6f;

			diagram.DiagramStyle.ShadowBrush = new SolidBrush(Color.FromRgba(0, 0, 0, 0x6E));
			diagram.DiagramStyle.FontFamily = "Arial";
			diagram.DiagramStyle.FontSize = 3f;
			diagram.ShadowOffsetX = 2;
			diagram.ShadowOffsetY = 2;
			diagram.DiagramLinkStyle.Brush = new SolidBrush(Color.FromRgb(0x41, 0x69, 0xE1));
			diagram.DiagramLinkStyle.ShadowBrush = new SolidBrush(Color.Black);
			diagram.DiagramLinkStyle.TextBrush = new SolidBrush(Color.FromRgb(0x00, 0x00, 0x8B));

			diagram.DiagramLinkStyle.FontStyle = FontAttributes.Bold;
			diagram.LinkHeadShape = ArrowHeads.Triangle;
			diagram.LinkHeadShapeSize = 3.5;
			diagram.ShapeNodeStyle.Effects.Add(new GlassEffect());

			diagram.Invalidate();
		}

		private void OnLoadClick(object sender, EventArgs e)
		{
			var assembly = typeof(App).GetTypeInfo().Assembly;
			Stream stream = assembly.GetManifestResourceStream("DecisionLayout.flowchart.xml");

			string text;
			using (var reader = new StreamReader(stream))
			{
				text = reader.ReadToEnd();
			}

			// Load the graph xml
			XDocument document = XDocument.Parse(text);

			diagram.LoadFromXml(document);
		}

		private void OnLayoutClick(object sender, EventArgs e)
		{
			Arrange();
		}

		private void OnRandomClick(object sender, EventArgs e)
		{
			diagram.ClearAll();

			var startNode = diagram.Factory.CreateShapeNode(0, 0, 30, 10);
			startNode.Brush = terminalBrush;
			startNode.Pen = terminalPen;
			startNode.Text = "START";
			startNode.Tag = "start";
			startNode.Shape = Shapes.Ellipse;

			var endNode = diagram.Factory.CreateShapeNode(0, 0, 30, 10);
			endNode.Brush = terminalBrush;
			endNode.Pen = terminalPen;
			endNode.Text = "END";
			endNode.Shape = Shapes.Ellipse;

			const int maxLength = 10;
			var all = new List<ShapeNode>();
			var remaining = new Queue<List<ShapeNode>>();

			remaining.Enqueue(new List<ShapeNode>(new[] { startNode }));

			var random = new Random(DateTime.Now.Millisecond);
			int length = 0;
			while (remaining.Count > 0)
			{
				var newNodes = new List<ShapeNode>();
				bool createdNew = false;

				var next = remaining.Dequeue();
				for (var n = 0; n < next.Count; n++)
				{
					var node = next[n];
					var isDecision = node.Tag is bool;

					ShapeNode previousChoice = null;
					int outLinks = isDecision ? 2 : 1;
					string[] linkTexts = isDecision ? new[] { "YES", "NO" } : new[] { "" };
					for (int i = 0; i < outLinks; i++)
					{
						// Create new or link to an existing node;
						// Create at least one new node for each level
						bool createNew = random.Next(maxLength) > length;
						if (!createdNew && n == next.Count - 1)
							createNew = true;
						if (length == maxLength)
							createNew = false;

						DiagramLink link;
						if (createNew)
						{
							// New decision or an operation node
							bool newDecision = random.Next(3) > 0 && !isDecision;
							var newNode = diagram.Factory.CreateShapeNode(0, 0, 30, 10);
							if (newDecision)
							{
								newNode.Shape = Shapes.Decision;
								newNode.AnchorPattern = AnchorPattern.Decision2In2Out;
								newNode.Brush = decisionBrush;
								newNode.Text = "Decision";
								newNode.Tag = true;
							}
							else
							{
								newNode.Shape = Shapes.Rectangle;
								newNode.Brush = normalBrush;
								newNode.Text = "Operation";
							}

							link = diagram.Factory.CreateDiagramLink(node, newNode);
							newNodes.Add(newNode);
							createdNew = true;
						}
						else
						{
							// Link to an existing node. If length == maxLength, link to the end node
							if (length == maxLength || remaining.Count == 0)
							{
								link = diagram.Factory.CreateDiagramLink(node, endNode);
							}
							else
							{
								// Make sure both choices don't lead to the same node
								ShapeNode choice = null;
								while (choice == previousChoice)
									choice = all[random.Next(all.Count)];
								link = diagram.Factory.CreateDiagramLink(node, choice);
								previousChoice = choice;
							}
						}

						link.Text = linkTexts[i];
					}

					all.Add(node);
				}

				if (newNodes.Count > 0)
					remaining.Enqueue(newNodes);

				length++;
			}

			Arrange();
		}

		void Arrange()
		{
			var layout = new MindFusion.Diagramming.Layout.DecisionLayout();
			layout.HorizontalPadding = 10;
			layout.VerticalPadding = 10;
			layout.StartNode = diagram.FindNode("start");
			layout.Arrange(diagram);
		}

		private Diagram diagram;
		Brush terminalBrush;
		Brush decisionBrush;
		Brush normalBrush;
		Pen terminalPen;
	}
}
